@@ -34,6 +34,9 @@ mongoc_uri_append_host (mongoc_uri_t *uri, const char *host, uint16_t port);
 bool
 mongoc_uri_parse_host (mongoc_uri_t *uri, const char *str);
 
+int32_t
+mongoc_uri_get_local_threshold_option (const mongoc_uri_t *uri);
+
 BSON_END_DECLS
 
 
